//=================================================================
//                    PluginLoader Class
//=================================================================
// 
//    This class is used to cycle through the jar/zip files located
// in the plugin directory under the framework's installation directory
// lib to find Plugin Modules. Only classes that match the interface
// Plugin_Module will be loaded.
//
//                     << PluginLoader.java >>
//
//=================================================================
// Copyright (C) 2016-2024 Dana M. Proctor
// Version 2.2 02/14/2024
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version
// 2 of the License, or (at your option) any later version. This
// program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty
// of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
// the GNU General Public License for more details. You should
// have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
// (http://opensource.org)
//
//=================================================================
// Revision History
// Changes to the code should be documented here and reflected
// in the present version number. Author information should
// also be included with the original copyright author.
//=================================================================
// Version 1.0 09/19/2016 Production PluginLoader Class.
//         1.1 09/24/2016 Updated References to PluginModule to Plugin_Module.
//         1.2 02/02/2017 Method cacheJAR() Added Additional Conditional to Handle Https
//                        URLs Separate From Http.
//         1.3 02/02/2017 Added Class Instance repositoryType & Used as Argument in Main
//                        Constructor & init(). Comment Changes for loadPluginEntry() in
//                        Addition to Breaking Out HTML jarFile Creation Into New Method
//                        loadHTMLPluginEntry(). Added Conditional for FTP(S) jarFile Creation
//                        Through New Method loadFTPPluginEntry().
//         1.4 02/02/2017 Added Class Instance repositoryOptions. Added Same to Main Constructor
//                        & init() Method. Method loadPluginEntry() Added Processing for
//                        Repository Types FTP(S). Removed Method loadHTMLPluginEntry() &
//                        loadFTPPluginEntry(). Modified Method cacheJAR() to Handle FTP(S)
//                        Repositories. Contained cacheJAR() Exceptions.
//         1.5 07/30/2019 Organized Imports. Removed Inner Class PriviledgedControl, Promoted
//                        to Regular Class in plugin Package. Substituted Call in loadPluginModule()
//                        for Change. Minor Formatting Changes.
//         1.6 12/19/2019 Method cacheJAR() Added Instances responseCode & responseMessage
//                        & Promoted httpConnection & httpsConnection to Method Instances
//                        so They Could be disconnected() in finally. Had to Combine HTTP
//                        or HTTPS Connection Creation in Same Method to Not Exceed Code
//                        Methods Bytes Limit.
//         1.7 12/20/2019 Method cacheJAR() Simplified Since HttpsURLConnection Extends
//                        HttpURLConnection, Removed Method Instance httpsURLConnection,
//                        responseCode, & responseMessage.
//         1.8 07/02/2020 Added Class Instance debug.
//         1.9 10/06/2023 Constructor Replaced URL(String), Deprecated With URI(String).toURL().
//                        Same in Methods loadPluginEntry(), cacheJAR(), and loadDefaultPlugin
//                        Entries(). Method loadPluginModules() Replaced AccessController.
//                        doPriviledged() With Direct Use of PriviledgedControl Then Calling
//                        run() to Create classLoader. Also in loadPluginModules() Replaced
//                        Deprecated module.newInstance() With module.getDeclaredConstuctor().
//                        newInstance().
//         2.0 02/07/2024 Method loadPluginEntry() Formatting Change, One Instance Per Line.
//                        Method init() Conditional for Deriving pluginFileName Replaced
//                        fileSeparator With Forward Slash. URI Creation Throughout Insured
//                        Proper Separator, Forward Slash.
//         2.1 02/10/2024 Method loadPluginModule() priviledgedControl Instantiation Changed
//                        Argument to True, Indicating this is a PluginModule.
//         2.2 02/10/2024 Method cacheJAR() Consolidated catch, IO/MalformedURL/URIExeception.
//                        
//-----------------------------------------------------------------
//                 danap@dandymadeproductions.com
//=================================================================

package com.dandymadeproductions.ajqvue.plugin;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.JarURLConnection;
import java.net.MalformedURLException;
import java.net.Proxy;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.net.ssl.HttpsURLConnection;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPReply;

import com.dandymadeproductions.ajqvue.Ajqvue;
import com.dandymadeproductions.ajqvue.gui.Main_Frame;
import com.dandymadeproductions.ajqvue.gui.PluginFrame;
import com.dandymadeproductions.ajqvue.io.WriteDataFile;
import com.dandymadeproductions.ajqvue.structures.GeneralProperties;
import com.dandymadeproductions.ajqvue.utilities.Utils;

/**
 *    The PluginLoader class is used to cycle through the jar/zip files
 * located in the plugin directory under the framework's installation
 * directory lib to find Plugin Modules. Only classes that match the
 * interface Plugin_Module will be loaded.
 * 
 * @author Dana M. Proctor
 * @version 2.2 02/14/2024
 */

public class PluginLoader implements Runnable
{
   // Class Instances.
   Thread pluginLoaderThread;
   private Main_Frame parentFrame;
   private URL pluginURL;
   private String repositoryName;
   private String repositoryType;
   private String[] repositoryOptions;
   
   private String fileSeparator;
   private String pluginDirectoryString;
   private String pluginFileName;
   
   private String pluginConfigFileString;
   private HashMap<String, String> pluginEntriesHashMap;
   
   private GeneralProperties generalProperties;
   
   private boolean debug;
   
   private static final String CONFIGURATION_FILENAME = "ajqvue_plugin.conf";
   private static final String VALID_PLUGIN_MODULENAME = "PluginModule.class";
   public static final String pathClassSeparator = "<$$$>";

   //==============================================================
   // PluginLoader Constructor(s)
   //
   // One Argument Constructor is used by the framework during
   // startup to load default plugins that installed from the
   // lib/plugins directory.
   //
   // Two Argument Constructor should be used to either load plugins
   // from the Plugin Management tool, or possibly externally by
   // outside classes.
   //==============================================================

   public PluginLoader(Main_Frame parent)
   {
      try
      {
         debug = Ajqvue.getDebug();
         
         init(parent, (new URI("file:/")).toURL(), "", PluginRepository.FILE, null);
      }
      catch (URISyntaxException | MalformedURLException mfe)
      {
         displayErrors("PluginLoader Constructor failed to create empty file:URL\n"
                       + mfe.toString());
      }
   }
   
   public PluginLoader(Main_Frame parent, URL pluginURL, String repositoryName, String repositoryType,
                       String[] repositoryOptions)
   {
      debug = Ajqvue.getDebug();
      
      init(parent, pluginURL, repositoryName, repositoryType, repositoryOptions);
   }
   
   //==============================================================
   // Class method for initialization and starting the class thread.
   //==============================================================
   
   private void init(Main_Frame parent, URL pluginURL, String repositoryName, String repositoryType,
                     String[] repositoryOptions)
   {
      parentFrame = parent;
      this.pluginURL = pluginURL;
      this.repositoryName = repositoryName;
      this.repositoryType = repositoryType;
      this.repositoryOptions = repositoryOptions;
      
      if (pluginURL == null)
         return;
      
      // Setup.
      fileSeparator = Utils.getFileSeparator();
      pluginConfigFileString = Utils.getAjqvueConfDirectory()
                               + Utils.getFileSeparator() + CONFIGURATION_FILENAME;
      generalProperties = Ajqvue.getGeneralProperties();
      
      pluginFileName = pluginURL.getFile();
      
      if (pluginFileName.indexOf("/") != -1)
         pluginFileName = pluginFileName.substring(pluginFileName.lastIndexOf("/") + 1,
                                                   pluginFileName.length());
      // System.out.println("PluginLoader init() pluginFileName: '" + pluginFileName + "'");
      
      // Default install directory lib/plugins
      if (pluginFileName.equals(""))
         pluginDirectoryString = Utils.getAjqvueDirectory() + fileSeparator
                                 + "lib" + fileSeparator + "plugins" + fileSeparator;
      // Specified plugin
      else
         pluginDirectoryString = "";
         
      pluginEntriesHashMap = new HashMap<String, String>();
      
      // Create and start the class thread.
      pluginLoaderThread = new Thread(this, "PluginLoader Thread");
      // System.out.println("PluginLoader Thread");
      
      pluginLoaderThread.start();
   }

   //==============================================================
   // Class method for normal start of the thread.
   //==============================================================

   public void run()
   {
      // Obtain the plugin module(s) & install into application.
      
      // Plugin Management Tool Load
      if (!pluginFileName.equals(""))
      {
         try
         {
            loadPluginEntry();
         }
         catch (IOException ioe){};
      }
      
      // Default lib/plugins and configuration load.
      else
      {
         try
         {
            loadDefaultPluginEntries();
            loadConfigurationFilePluginEntries();
         }
         catch (IOException ioe)
         {
            displayErrors("PluginLoader run() \n" + ioe.toString());
         }
      }
      loadPluginModules();
   }
   
   //==============================================================
   // Class Method for loading a plugin module manually through
   // the framework Plugin Management tool in the top tab.
   //
   // File URL - file:lib/plugins/TableRecordCount.jar
   // Http URL - http(s)://dandymadeproductions.com/temp/TableRecordCount.jar
   // JAR URL - jar:file:/home/duke/duke.jar!
   // FTP URL - ftp(s):/dandymadeproductions.com/
   //
   // Other Valid Java URL protocols, gopher, mailto, appletresource,
   // doc, netdoc, systemresource, & verbatim.
   //==============================================================

   private void loadPluginEntry() throws IOException
   {
      // Method Instances
      String className;
      String currentFileName;
      String pluginEntry;
      URL loadingPluginURL;
      
      ZipFile jarFile;
      File configurationFile;
      FileWriter fileWriter;
      
      // Check for a a valid jar file to be processed
      // then search for a plugin module.
      
      if (!pluginFileName.toLowerCase(Locale.ENGLISH).endsWith(".jar"))
         return;
      
      loadingPluginURL = pluginURL;
      jarFile = null;
      fileWriter = null;
      
      try
      {
         // Create a URL & then file to the JAR file
         // so that it can be searched.
         
         // Local File system plugin
         if (pluginURL.getProtocol().equals(PluginRepository.FILE))
         {
            jarFile = new ZipFile(new File((pluginURL.toExternalForm()).substring(
               pluginURL.toExternalForm().indexOf("file:") + 5)));
         }
         
         // Http(s)/Ftp(s) plugin
         else if (pluginURL.getProtocol().startsWith(PluginRepository.HTTP)
                  || pluginURL.getProtocol().startsWith(PluginRepository.FTP))
         { 
            Proxy httpProxy = null;
            String cachedJAR_FileName;
            
            // Try proxy
            if (generalProperties.getEnableProxy() && repositoryType.startsWith(PluginRepository.HTTP))
               httpProxy = new Proxy(Proxy.Type.HTTP, new InetSocketAddress(
                  generalProperties.getProxyAddress(), generalProperties.getProxyPort()));
            
            // Try to Cache, all remote network plugins MUST
            // be cached!
            cachedJAR_FileName = Utils.getCacheDirectory() + repositoryName
                                 + fileSeparator + pluginFileName;
               
            if (cacheJAR(pluginURL.toExternalForm(), cachedJAR_FileName, httpProxy, true))
            {
               loadingPluginURL = (new URI(
                  "file:" + cachedJAR_FileName.replace("\\", "/"))).toURL();
               if (debug)
                  System.out.println("PluginLoader loadPluginEntry() JAR Cached: " + cachedJAR_FileName);
            }
            else
            {
               JOptionPane.showMessageDialog(null, "PluginLoader loadPluginEntry() "
                                             + "Failed Cache of Network JAR", "Alert",
                                             JOptionPane.ERROR_MESSAGE);
               return;
            }
            
            // Proceed to use the cached plugin.
            URL jarUrl = (new URI("jar:" + loadingPluginURL.toExternalForm() + "!/")).toURL();
            JarURLConnection conn = (JarURLConnection) jarUrl.openConnection();
            jarFile = conn.getJarFile();
         }
         
         // Unknown
         else
            return;
          
         // Search
         className = "";
         
         for (Enumeration<?> entries = jarFile.entries(); entries.hasMoreElements();)
         {
            currentFileName = ((ZipEntry) entries.nextElement()).getName();

            // Plugin Qualifier
            if (currentFileName.endsWith(".class") && currentFileName.indexOf("$") == -1
                && currentFileName.indexOf(VALID_PLUGIN_MODULENAME) != -1)
            {
               currentFileName = currentFileName.replaceAll("/", ".");
               currentFileName = currentFileName.substring(0, currentFileName.indexOf(".class"));
               className = currentFileName;
               
               if (className.startsWith("java.") || className.startsWith("javax."))
                  continue;
               
               pluginEntriesHashMap.put(loadingPluginURL.toExternalForm(), className);
               // System.out.println("PluginLoader loadPluginEntry() Located: "
               //                     + loadingPluginURL.toExternalForm() + " " + className);
            }
         }
         
         // Update the configuration file indicating valid
         // plugin modules that have been loaded manually.
         
         if (!pluginFileName.equals("") && !className.equals(""))
         {
            pluginEntry = loadingPluginURL.toExternalForm() + pathClassSeparator + className + "\n";
            
            // Write new or appending. 
            configurationFile = new File(pluginConfigFileString);
               
            if (!configurationFile.exists())
               WriteDataFile.mainWriteDataString(pluginConfigFileString, pluginEntry.getBytes(), false);
            else
            {
               fileWriter = new FileWriter(pluginConfigFileString, true);
               char[] buffer = new char[pluginEntry.length()];
               pluginEntry.getChars(0, pluginEntry.length(), buffer, 0);
               fileWriter.write(buffer);
               fileWriter.flush();
            }
         }
      }
      catch (URISyntaxException | MalformedURLException e)
      {
         displayErrors("PluginLoader loadPluginEntry() MalFormedException: " + pluginFileName + "\n"
                        + e.toString());
      }
      catch (IOException e)
      {
         displayErrors("PluginLoader loadPluginEntry() IOException: " + pluginFileName + "\n"
                        + e.toString());
      }
      finally
      {
         try
         {
            if (jarFile != null)
               jarFile.close();
         }
         catch (IOException ioe)
         {
            displayErrors("PluginLoader loadPluginEntry() Failed to close jarFile\n"
                          + ioe.toString());
         }
         finally
         {
            if (fileWriter != null)
               fileWriter.close();
         }
      }
   }
   
   //==============================================================
   // Class Method to cache a JAR file if it is not a local file
   // system resource.
   //==============================================================

   private boolean cacheJAR(String resourceURLString, String cachedJAR_FileName,
                            Proxy proxy, boolean debugMode)
   {
      // Method Instances
      HttpURLConnection httpConnection;
      FTPClient ftpClient;
      InputStream inputStream;
      BufferedInputStream inputBuffer;
      int inByte;

      File cacheDirectoryFile;
      FileOutputStream fileOutputStream;
      BufferedOutputStream fileOutputBuffer;
      
      boolean cached;

      // Setup
      httpConnection = null;
      ftpClient = null;
      inputStream = null;
      inputBuffer = null;
      fileOutputStream = null;
      fileOutputBuffer = null;
      
      cached = false;

      // Open a stream for the JAR then byte it to the
      // local cache file.
      try
      {
         URL url = (new URI(resourceURLString.replace("\\", "/"))).toURL();
         
         // Handle HTTP/HTTPS
         if (pluginURL.getProtocol().startsWith(PluginRepository.HTTP))
         {
            if (proxy != null)
            {
               if (pluginURL.getProtocol().equals(PluginRepository.HTTP))
                  httpConnection = (HttpURLConnection) url.openConnection(proxy);
               else
                  httpConnection = (HttpsURLConnection) url.openConnection(proxy);
            }
            else
            {
               if (pluginURL.getProtocol().equals(PluginRepository.HTTP))
                  httpConnection = (HttpURLConnection) url.openConnection(Proxy.NO_PROXY);
               else
                  httpConnection = (HttpsURLConnection) url.openConnection(Proxy.NO_PROXY);
            }
            
            // Authorization Needed?
            if (httpConnection.getResponseCode() != HttpURLConnection.HTTP_OK)
            {
               displayErrors("PluginLoader cacheJAR()\n"
                             + httpConnection.getResponseMessage());
               return cached;
            }
            else
               inputStream = httpConnection.getInputStream();
         }
         // Handle FTP(S)
         else if (pluginURL.getProtocol().startsWith(PluginRepository.FTP))
         {  
            FTP_Client ftp_Client;
            
            if (repositoryType.equals(PluginRepository.FTPS))
               ftp_Client = new FTP_Client(repositoryType, PluginRepository.FTPS
                                           + resourceURLString.substring(resourceURLString.indexOf(":")),
                                           repositoryOptions);
            else
               ftp_Client = new FTP_Client(repositoryType, resourceURLString, repositoryOptions);
            
            ftpClient = ftp_Client.createFTPClient();
            
            if (ftpClient == null)
               return cached;
            
            fileOutputStream = new FileOutputStream(cachedJAR_FileName);
            
            ftpClient.retrieveFile(pluginFileName, fileOutputStream);
            
            if (ftpClient.getReplyCode() == FTPReply.FILE_UNAVAILABLE)
            {
               displayErrors("PluginLoader cacheJAR() Plugin Not Found:\n"
                             + ftpClient.getReplyCode() + " : "
                             + ftpClient.getReplyString());
            }
            else
               cached = true;
            
            ftpClient.noop();
            ftpClient.logout();
            
            return cached;   
         }
         // Handle File
         else
            inputStream = ((new URI(
               resourceURLString.replace("\\", "/")))).toURL().openStream();
         
         inputBuffer = new BufferedInputStream(inputStream);

         cacheDirectoryFile = new File(Utils.getCacheDirectory());

         if (!cacheDirectoryFile.isDirectory())
            if (!cacheDirectoryFile.mkdir())
               if (debugMode)
               {
                  System.out.println("PluginLoader" + " cacheJAR()\n"
                                     + "Failed to create Cache Directory");
                  return cached;
               }

         fileOutputStream = new FileOutputStream(cachedJAR_FileName);
         fileOutputBuffer = new BufferedOutputStream(fileOutputStream);

         while ((inByte = inputBuffer.read()) != -1)
            fileOutputBuffer.write(inByte);

         fileOutputBuffer.flush();
         cached = true;
      }
      catch (URISyntaxException | IOException  e)
      {
         if (e instanceof IOException && debug)
            System.out.println("PluginLoader" + " cacheJAR()\n"
                               + "Failed to Cache JAR.\n" + e.toString());
         else
            displayErrors("PluginLoader cacheJAR() MalformedURL/URIException: "
                          + pluginFileName + "\n" + e.toString());
      }
      finally
      {
         try
         {
            if (fileOutputBuffer != null)
               fileOutputBuffer.close();
         }
         catch (IOException ioe)
         {
            if (debugMode)
               System.out.println("PluginLoader" + " cacheJAR() Failed to Close "
                                  + "fileOutputBuffer.\n" + ioe.toString());
         }
         finally
         {
            try
            {
               if (fileOutputStream != null)
                  fileOutputStream.close();
            }
            catch (IOException ioe)
            {
               if (debugMode)
                  System.out.println("PluginLoader" + " cacheJAR() Failed to Close "
                                     + "fileOutputStream.\n" + ioe.toString());
            }
            finally
            {
               try
               {
                  if (inputBuffer != null)
                     inputBuffer.close();
               }
               catch (IOException ioe)
               {
                  if (debugMode)
                     System.out.println("PluginLoader" + " cacheJAR() Failed to Close "
                                        + "inputBuffer.\n" + ioe.toString());
               }
               finally
               {
                  try
                  {
                     if (inputStream != null)
                        inputStream.close();  
                  }
                  catch (IOException ioe)
                  {
                     if (debugMode)
                        System.out.println("PluginLoader cacheJAR()() Failed to Close "
                                           + "inputStream.\n" + ioe.toString());
                  }
                  finally
                  {
                     try
                     {
                        if (httpConnection != null)
                           httpConnection.disconnect();
                        
                        if (ftpClient != null && ftpClient.isConnected())
                           ftpClient.disconnect();
                     }
                     catch (IOException ioe)
                     {
                        if (debugMode)
                           System.out.println("PluginLoader cacheJAR() "
                                              + "Failed to disconnet ftpClient. " + ioe.toString());
                     }     
                  }
               }
            }
         }    
      }
      return cached;
   }

   //==============================================================
   // Class Method for reviewing the JAR files in the lib/plugin/
   // directory in search of a PluginModule to be then loaded.
   //==============================================================

   private void loadDefaultPluginEntries() throws IOException
   {
      // Method Instances
      File filePluginsDirectory;
      String[] jarFileNames;
      URL defaultEntriesURL;
      String pathKey;
      String currentFileName;
      ZipFile jarFile;

      // Create the file for the plugin directory & load
      // directory contents.

      filePluginsDirectory = new File(pluginDirectoryString);
      
      if (!filePluginsDirectory.exists())
         return;
      
      jarFileNames = filePluginsDirectory.list();
      
      if (jarFileNames == null)
         return;
      
      jarFile = null;

      // Cycle through the files in search of plugins.
      for (int i = 0; i < jarFileNames.length; i++)
      {
         if (!jarFileNames[i].toLowerCase(Locale.ENGLISH).endsWith(".jar"))
            continue;

         try
         {
            jarFile = new ZipFile(pluginDirectoryString + jarFileNames[i]);

            for (Enumeration<?> entries = jarFile.entries(); entries.hasMoreElements();)
            {
               currentFileName = ((ZipEntry) entries.nextElement()).getName();

               // Plugin Qualifier
               if (currentFileName.endsWith(".class") && currentFileName.indexOf("$") == -1
                   && currentFileName.indexOf(VALID_PLUGIN_MODULENAME) != -1)
               {
                  defaultEntriesURL = (new URI("file:"
                                               + pluginDirectoryString.replace("\\", "/")
                                               + jarFileNames[i])).toURL();
                  pathKey = defaultEntriesURL.toExternalForm();
                  // System.out.println("PluginLoader loadDefaultPluginEntries() Located:" + pathKey);
                  
                  currentFileName = currentFileName.replaceAll("/", ".");
                  currentFileName = currentFileName.substring(0, currentFileName.indexOf(".class"));
                  
                  if (currentFileName.startsWith("java.") || currentFileName.startsWith("javax."))
                     continue;
                  
                  pluginEntriesHashMap.put(pathKey, currentFileName);
               }
            }
         }
         catch (URISyntaxException | MalformedURLException e)
         {
            displayErrors("PluginLoader loadDefaultPluginEntries() MalformedURLException: "
                          + jarFileNames[i] + "\n" + e.toString());
         }
         catch (IOException e)
         {
            displayErrors("PluginLoader loadDefaultPluginEntries() IOException: "
                  + jarFileNames[i] + "\n" + e.toString());
         }
         finally
         {
            if (jarFile != null)
               jarFile.close();
         }
      }
   }
   
   //==============================================================
   // Class Method for reviewing the entries in the configuration
   // file ajqvue_plugin.conf file to be loaded as plugins.
   //==============================================================

   private void loadConfigurationFilePluginEntries() throws IOException
   {
      // Method Instances
      String currentLine, pathKey, className;
      File configurationFile;
      FileReader fileReader;
      BufferedReader bufferedReader;
      
      fileReader = null;
      bufferedReader = null;
      
      try
      {
         // Check to see if file exists.
         configurationFile = new File(pluginConfigFileString);
         try
         { 
            if (!configurationFile.exists())
               return;
         }
         catch (SecurityException e)
         {
            displayErrors("PluginLoader loadConfigurationFilePluginEntries() Security Exception: "
                          + e.toString());
            return;
         }
         
         // Looks like there is a plugin configuration file
         // so collect the entries.
         
         fileReader = new FileReader(pluginConfigFileString);
         bufferedReader = new BufferedReader(fileReader);
            
         while ((currentLine = bufferedReader.readLine()) != null)
         {
            currentLine = currentLine.trim();
            
            if (currentLine.indexOf(pathClassSeparator) != -1)
            {
               pathKey = currentLine.substring(0, currentLine.indexOf(pathClassSeparator));
               className = currentLine.substring(currentLine.indexOf(pathClassSeparator)
                                                 + pathClassSeparator.length());
               
               if (className.startsWith("java.") || className.startsWith("javax."))
                  continue;
              
               pluginEntriesHashMap.put(pathKey, className);
               // System.out.println("PluginLoader loadConfigurationFilePluginEntry() Located:" + pathKey); 
            }
            else
               continue;
         }
      }
      catch (IOException ioe) 
      {
         displayErrors("PluginLoader loadConfigurationFilePluginEntries() File I/O problem, reading "
                       + pluginConfigFileString + "\n" + ioe.toString());
      }
      finally
      {
         try
         {
            if (bufferedReader != null)
               bufferedReader.close();
         }
         catch (IOException ioe)
         {
            displayErrors("PluginLoader loadConfigurationFilePluginEntries() Failed to close bufferReader\n"
                          +ioe.toString());
         }
         finally
         {
            if (fileReader != null)
               fileReader.close();
         }
      }
   }

   //==============================================================
   // Class Method for loading the actual plugin module class
   // instances
   //==============================================================

   private void loadPluginModules()
   {
      // Method Instances.
      URLClassLoader classLoader;
      PriviledgedControl priviledgedControl;
      
      String iconsDirectory;
      ImageIcon defaultModuleIcon;

      // Obtain & create default Image Icon.

      iconsDirectory = Utils.getIconsDirectory() + fileSeparator;
      defaultModuleIcon = Ajqvue.getResourceBundle().getResourceImage(iconsDirectory
                                                                          + "newsiteLeafIcon.png");

      // Iterator through the found plugins and load them.

      Set<Map.Entry<String, String>> keySet = pluginEntriesHashMap.entrySet();
      Iterator<Map.Entry<String, String>> pluginIterator = keySet.iterator();

      while (pluginIterator.hasNext())
      {
         Map.Entry<String, String> pluginEntry = pluginIterator.next();
         
         final String pluginURLString = pluginEntry.getKey();
         // System.out.println("PluginLoader loadPluginModules() pluginURLString: " + pluginURLString);
         
         priviledgedControl = new PriviledgedControl(pluginURLString, true, debug);
         classLoader = (URLClassLoader) priviledgedControl.run();
         
         // If looks like a good plugin try to load it.
         
         if (classLoader != null)
         {
            // Create the instance and add to the framework.
            try
            {
               if (debug)
                  System.out.println("PluginLoader loadPluginModules() "
                                     + pluginEntry.getValue());
               try
               {
                  Class<?> module = Class.forName(pluginEntry.getValue(), true, classLoader);
                  
                  if (module.getDeclaredConstructor().newInstance() instanceof Plugin_Module)
                  {
                     Plugin_Module pluginModule = (Plugin_Module) module.getDeclaredConstructor().newInstance();
                     pluginModule.path_FileName = pluginEntry.getKey() + pathClassSeparator + pluginEntry.getValue();

                     new PluginThread(parentFrame, pluginModule, defaultModuleIcon);
                  }
                  else
                     throw new Exception();
               }
               catch (ClassNotFoundException cnfe)
               {
                  throw new Exception(cnfe);
               }
               catch (NoClassDefFoundError ncdfe)
               {
                  throw new Exception(ncdfe);
               }
            }
            catch (Exception e)
            {
               displayErrors("PluginLoader loadPluginModules() Exception: \n" + e.toString());
               Thread removePluginConfigurationModuleThread = new Thread(new Runnable()
               {
                  public void run()
                  {
                     PluginFrame.removePluginConfigurationModule(pluginURLString);
                  }
               }, "PluginLoader.removePluginConfigurationModuleThread2");
               removePluginConfigurationModuleThread.start();
            }
         }
      }
   }
   
   //==============================================================
   // Class method to display an error to the standard output if
   // the debug option is active.
   //==============================================================

   public void displayErrors(String errorString)
   {
      if (debug)
      {
         System.out.println(errorString);
      }
   }
}
